/*
 *
 */
import java.awt.EventQueue;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.List;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.logging.Logger;
import javax.swing.JOptionPane;


/*
 *
 */
import installer.common.Controller;
import installer.common.InstallerGUI;
import installer.common.NoSuperuserPasswordException;
import installer.common.Pointer;
import installer.common.ProcessExecutionException;
import installer.common.ProcessOutputAnalyzer;
import installer.common.ProcessRunner;


/*
 *
 */
class NativeController extends Controller
{
    /*
     *
     */
    // static final String nativeBinaryListFileName = "native-lib-files.txt";
    static final String nativeX86BinaryListFileName = "native-lib-files-x86.txt";
    static final String nativeAMD64BinaryListFileName = "native-lib-files-amd64.txt";


    // /*
    //  *
    //  */
    // private Logger logger;


    /*
     *
     */
    public NativeController()
    {
        // this.logger = null;
    }


    /*
     * FIXME: run as administrator if failed!
     *
     * FIXME: this assumes (because xcopy forces a prompt otherwise) that the
     *        two filenames are the same, and just located in different directories.
     *        ugh.
     */
    protected void copyFile(final File sourceFile,
                            final File destinationFile)
        throws IOException,
               NoSuperuserPasswordException
    {
        ProcessRunner.Result<?> xcopyResult;
        boolean copiedFile;
        String[] commandLine =
            { "xcopy", "/Y",
              sourceFile.getAbsolutePath(),
              (destinationFile.exists()
               ? destinationFile.getAbsolutePath()
               : destinationFile.getParent()) };


        // FIXME: this is in lieu of asking for the administrator password
        while (true)
        {
            try
            {
                xcopyResult = ProcessRunner.run(null, commandLine);
            }
            catch (ProcessExecutionException ex)
            {
                copiedFile = false;
                // FIXME: log this
                // FIXME: should this really always be turned into an IOException?
                throw new IOException(ex.getMessage());
            }

            if (xcopyResult.processReturnResult == 0)   // xcopy succeeded
            {
                copiedFile = true;
                this.logger.info("Copied " + sourceFile.getAbsolutePath() + " to "
                                 + destinationFile.getAbsolutePath() + " successfully.\n");
                break;
            }
            else   // xcopy failed
            {

                final Pointer<Boolean> okToTryAgain = new Pointer<Boolean>(true);

                //
                this.logger.info("Tried to copy " + sourceFile.getAbsolutePath() + " to "
                                 + destinationFile.getAbsolutePath() + ", but failed.\n");

                // FIXMEFIXME: it's really a hack to have this here; it should be abstracted
                try
                {
                    EventQueue.invokeAndWait(new Runnable()
                        {
                            public void run()
                            {
                                int result;

                                // FIXME: use frame as first arg.
                                result = 
                                    JOptionPane.showConfirmDialog(null,
                                                                  "The installation needs to copy \n"
                                                                  + sourceFile.getAbsolutePath() + " \n"
                                                                  + "to "
                                                                  + destinationFile.getAbsolutePath()
                                                                  + "\n"
                                                                  + "but is not able to do so; this is "
                                                                  + "most likely because of a lack of "
                                                                  + "proper permission.\n"
                                                                  + "Please make the file accessible "
                                                                  + "(non-read-only, and writeable by "
                                                                  + "your user account) and then click "
                                                                  + "\"OK\" to try again.\n"
                                                                  + "If you choose to cancel, "
                                                                  + "installation for this version of "
                                                                  + "MATLAB will not be completed.",
                                                                  "Error Copying Files",
                                                                  JOptionPane.OK_CANCEL_OPTION);
                                okToTryAgain.object = (result == JOptionPane.OK_OPTION);
                            }
                        });

                    if (okToTryAgain.object)
                    {
                        continue;
                    }
                    else
                    {
                        copiedFile = false;
                        break;
                    }
                }
                catch (InterruptedException ex)
                {
                    copiedFile = false;
                    break;
                }
                catch (InvocationTargetException ex)
                {
                    copiedFile = false;
                    break;
                }
            }
        }

        if (!copiedFile)
        {
            throw new IOException("Unable to copy "
                                  + sourceFile.getAbsolutePath()
                                  + " to " + destinationFile.getAbsolutePath());
        }
    }


    /*
     *
     */
    public Map<File, InstallerGUI.MatlabInstallation> findMatlabPathsAndVersions()
    {
        List<String> shortVersionStrings;
        Map<File, InstallerGUI.MatlabInstallation> pathsToVersions;


        //
        pathsToVersions = new HashMap<File, InstallerGUI.MatlabInstallation>();

        //
        try
        {
            this.logger.info("Looking for MATLAB paths...\n");
            shortVersionStrings =
                ProcessRunner.run(new ProcessOutputAnalyzer.SingleGroupRegexpAll("^.*MATLAB\\\\(.*)$"),
                                  "reg", "query",
                                  "HKLM\\SOFTWARE\\MathWorks\\MATLAB").outputResult;
        }
        catch (ProcessExecutionException ex)
        {
            this.logger.severe("Unable to find MATLAB paths:\n");
            this.logger.severe(ex.getMessage() + "\n");

            // pathsToVersions is empty here
            return pathsToVersions;
        }

        //
        for (String shortVersionString : shortVersionStrings)
        {
            try
            {
                String pathString;
                File path;
                InstallerGUI.MatlabInstallation matlabInst;

                this.logger.info("Trying to get path for MATLAB version "
                                 + shortVersionString + " ...\n");

                pathString =
                    ProcessRunner.run(new ProcessOutputAnalyzer.SingleGroupRegexpFirst("^\\s*MATLABROOT\\s*REG_SZ\\s*(.*)$"),
                                      "reg", "query",
                                      "HKLM\\SOFTWARE\\MathWorks\\MATLAB\\" + shortVersionString,
                                      "/v", "MATLABROOT").outputResult;
                path = new File(pathString);
                
                matlabInst = verifyMatlabPathAndGetVersion(path);
                if (matlabInst == null)
                {
                    this.logger.severe("Unable to verify MATLAB installation in "
                                       + pathString + "\n");
                    continue;
                }

                this.logger.info("Got path for MATLAB version " + shortVersionString + ": "
                                 + pathString + "\n");

                pathsToVersions.put(path, matlabInst);
            }
            catch (ProcessExecutionException ex)
            {
                this.logger.severe("Unable to get MATLAB path:\n");
                this.logger.severe(ex.getMessage() + "\n");
                continue;
            }
        }

        //
        return pathsToVersions;
    }


    /*
     *
     */
    protected boolean
        gatherPlatformDependentConfigurationInformation(List<String> errors) // out
    {
        return true;
    }


    /*
     *
     */
    protected File getBinaryFilePathFromMatlabPath(File matlabPath)
    {
        File binaryPath;


        //
        binaryPath = super.getBinaryFilePathFromMatlabPath(matlabPath);
        if (binaryPath.exists())
        {
            return binaryPath;
        }

        //
        binaryPath = new File(new File(matlabPath, "bin"), "matlab.exe");
        if (binaryPath.exists())
        {
            return binaryPath;
        }

        // we don't test this, because we aren't expected to return null
        return new File(new File(matlabPath, "bin"), "matlab.bat");
    }


    /*
     *
     */
    protected File getMatlabUserpathGuess()
    {
        return new File(new File(getUserHomeDirectory(), "My Documents"), "MATLAB");
    }


    /*
     *
     */
    protected Set<String> getNativeBinaryFileNames(String arch)
        throws IOException
    {
        if (arch.equals("x86"))
        {
            return getResourceNamesFromTextFileResource(nativeX86BinaryListFileName);
        }
        else if (arch.equals("amd64"))
        {
            return getResourceNamesFromTextFileResource(nativeAMD64BinaryListFileName);
        }
        else
        {
            // this should never happen
            throw new IOException("Unsupported arch. libs. requested (this is a bug!)");
        }
    }


    /*
     *
     */
    protected boolean isFileReadOnly(File file)
        throws IOException,
               ProcessExecutionException
    {
        ProcessOutputAnalyzer.SingleGroupRegexpFirst attribAnalyzer;
        ProcessRunner.Result<String> attribResult;


        if (file.exists())
        {
            //
            attribAnalyzer = new ProcessOutputAnalyzer.SingleGroupRegexpFirst("^.....([ R]).*$");

            //
            attribResult =
                ProcessRunner.run(new ProcessOutputAnalyzer.FirstLineOnly<String>(attribAnalyzer),
                                  "attrib", file.getAbsolutePath());
        
            //
            return attribResult.outputResult.equals("R");
        }
        else
        {
            // FIXME: should probably return whether the parent dir. is read-only
            return false;
        }
    }


    /*
     * arch should be non-null
     */
    protected boolean isSupportedArch(String arch)
    {
        return arch.equals("x86") || arch.equals("amd64");
    }


    /*
     * version should be non-null
     *
     * FIXME FILLME
     */
    protected boolean isSupportedMatlabVersion(String version)
    {
        return true;
    }


    /*
     * matlabCommand must not contain any semi-colons or
     * whitespace (most likely it will be just a single
     * function call without arguments)
     */
    protected void launchMatlabWithCommand(File matlabPath,
                                           String matlabCommand,
                                           File workingDirectory)
        throws IOException
    {
        String workingDirString;
        String chdirAndMatlabCommand;


        // FIXME: this is a bit of a hack
        workingDirString = escapeStringForMatlab(workingDirectory.getAbsolutePath());
        chdirAndMatlabCommand = "chdir('" + workingDirString + "');" + matlabCommand;
        
        //
        try
        {
            WindowsCommandLineMatlabRunner.get().launch(matlabPath,
                                                        chdirAndMatlabCommand,
                                                        workingDirectory);
            return;
        }
        catch (Exception ex)
        {
            this.logger.warning("Failed to run MATLAB's exe file: " + ex.getMessage() + ";"
                                + "trying arch-specific exe file instead\n");
        }

        //
        try
        {
            WindowsArchSubdirMatlabRunner.get().launch(matlabPath,
                                                       chdirAndMatlabCommand,
                                                       workingDirectory);
            return;
        }
        catch (Exception ex)
        {
            this.logger.warning("Failed to run MATLAB's arch-specific exe file: "
                                + ex.getMessage() + "; "
                                + "trying bat file instead\n");
        }

        // FIXME: this is a last resort -- it uses a temp. dir. and temp. file, which is
        //        not cleaned up
        WindowsTempFileMatlabRunner.get().launch(matlabPath, matlabCommand, workingDirectory);
    }


    /*
     * FIXME: this is a bit of a hack
     *
     * we are overriding this because of an observed problem renaming the directory on windows;
     * FIXME: at the moment, the exact reason for the problem is unclear -- launching a command
     * line like "CMD /C REN SRC DST" would almost work, except we couldn't get a useful exit
     * code. a bit of native code is needed.
     */
    public void moveTempConfigDirToFinalConfigDir()
        throws IOException
    {
        boolean renamedDir;


        //
        renamedDir = false;
        while (!renamedDir)
        {
            try
            {
                super.moveTempConfigDirToFinalConfigDir();
                renamedDir = true;
                break;
            }
            catch (IOException ex)
            {
                int result;
                final Pointer<Boolean> okToTryAgain = new Pointer<Boolean>(true);

                //
                renamedDir = false;

                //
                this.logger.info("Tried to move " + getTempConfigDirectory().getAbsolutePath() + " to "
                                 + getConfigDirectory().getAbsolutePath() + ", but failed.\n");

                // FIXMEFIXME: it's really a hack to have this here; it should be abstracted
                try
                {
                    EventQueue.invokeAndWait(new Runnable()
                        {
                            public void run()
                            {
                                int result;

                                // FIXME: use frame as first arg.
                                result =
                                    JOptionPane.showConfirmDialog(null,
                                                                  "The installation needs to rename \n"
                                                                  + getTempConfigDirectory().getAbsolutePath() + " \n"
                                                                  + "to "
                                                                  + getConfigDirectory().getAbsolutePath() + "\n"
                                                                  + "but is not able to do so; this may be because of a\n"
                                                                  + "lack of proper permissions, or because some other program\n"
                                                                  + "is using the former directory (which would be very strange.)\n"
                                                                  + "Please ensure that the proper permissions are available, and\n"
                                                                  + "that no other program is using the directory, and then click\n"
                                                                  + "\"OK\" to try again. If you choose to cancel, installation\n"
                                                                  + "will not be able to continue.",
                                                                  "Error Renaming Temp. Directory",
                                                                  JOptionPane.OK_CANCEL_OPTION);
                                okToTryAgain.object = (result == JOptionPane.OK_OPTION);
                            }
                        });

                    if (okToTryAgain.object)
                    {
                        continue;
                    }
                    else
                    {
                        renamedDir = false;
                        break;
                    }
                }
                catch (InterruptedException ieEx)
                {
                    renamedDir = false;
                    break;
                }
                catch (InvocationTargetException iteEx)
                {
                    renamedDir = false;
                    break;
                }
            }
        }

        if (!renamedDir)
        {
            throw new IOException("Unable to rename " + getTempConfigDirectory().getAbsolutePath()
                                  + " to " + getConfigDirectory().getAbsolutePath());
        }
    }


    /*
     * the reader may notice that we create a log file no matter what,
     * as this is a guaranteed way to read the output on windows (if there
     * is another way, i don't know what it is yet)
     *
     * FIXME: the log messages here are not that helpful, probably
     */
    public <ResultType> ResultType
        runMatlabCommand(File matlabPath,
                         String matlabCommand,
                         File logFile,
                         ProcessOutputAnalyzer<ResultType> outputAnalyzer)
        throws IOException,
               ProcessExecutionException
    {
        try
        {
            return WindowsCommandLineMatlabRunner.get().run(matlabPath, matlabCommand,
                                                            logFile, outputAnalyzer);
        }
        catch (Exception ex)
        {
            this.logger.warning("Failed to run MATLAB's exe file: " + ex.getMessage() + "; "
                                + "trying arch-specific exe file instead\n");
        }

        try
        {
            return WindowsArchSubdirMatlabRunner.get().run(matlabPath, matlabCommand,
                                                           logFile, outputAnalyzer);
        }
        catch (Exception ex)
        {
            this.logger.warning("Failed to run MATLAB's arch-specific exe file: "
                                + ex.getMessage() + "; "
                                + "trying bat file instead\n");
        }

        return WindowsTempFileMatlabRunner.get().run(matlabPath, matlabCommand,
                                                     logFile, outputAnalyzer);

        // // DEBUG DEBUG DEBUG

        // try
        // {
        //     return WindowsTempFileMatlabRunner.get().run(matlabPath, matlabCommand,
        //                                                  logFile, outputAnalyzer);
        // }
        // catch (Exception ex)
        // {
        //     this.logger.warning("Failed to run MATLAB's bat file: " + ex.getMessage() + "\n");
        // }
        
        // return WindowsCommandLineMatlabRunner.get().run(matlabPath, matlabCommand,
        //                                                 logFile, outputAnalyzer);


    }

    
    /*
     *
     */
    protected void setFileReadOnly(File file,
                                   boolean readOnly)
        throws IOException,
               ProcessExecutionException
    {
        ProcessRunner.Result<?> attribResult;


        //
        attribResult =
            ProcessRunner.run(null, "attrib", (readOnly ? "+R" : "-R"), file.getAbsolutePath());

        // FIXMEFIXME: should throw a ProcessExecutionException on failure, but need to figure
        //             out what a failure return value looks like, first
    }


    // /*
    //  * FIXME: this is a bit brittle...
    //  */
    // public void setLogger(Logger logger)
    // {
    //     this.logger = logger;
    // }


    // /*
    //  *
    //  */
    // public InstallerGUI.MatlabInstallation verifyMatlabPathAndGetVersion(File matlabPath)
    //     throws IOException
    // {
    //     final String matlabVersionAndArchCommand =
    //         "nl=['\\','n'];fprintf(1,['%s',nl,'%s',nl], "
    //         + "version, "
    //         + "char(java.lang.System.getProperty('os.arch')));"
    //         + "exit;";
            

    //     //
    //     try
    //     {
    //         String arch;
    //         String version;
    //         List<String> fullOutput;

    //         //
    //         fullOutput = runMatlabCommand(matlabPath, matlabVersionAndArchCommand, null,
    //                                       new ProcessOutputAnalyzer.EveryLine());
    //         if (fullOutput == null
    //             || (fullOutput.size() < 2))
    //         {
    //             logger.severe("Unable to get version and/or arch. for MATLAB in "
    //                           + matlabPath.getAbsolutePath() + "\n");
    //             return null;
    //         }

    //         //
    //         version = fullOutput.get(fullOutput.size()-2);
    //         arch = fullOutput.get(fullOutput.size()-1);

    //         //
    //         if (arch == null
    //             || (!arch.equals("x86") && !arch.equals("amd64")))
    //         {
    //             logger.severe("Unable to use MATLAB in " + matlabPath.getAbsolutePath() + " -- "
    //                           + "the JVM's architecture is "
    //                           + ((arch == null) ? "unknown (Java property os.arch missing!)"
    //                              : arch) + ", "
    //                           + "but currently only x86 and amd64 are supported!\n");
    //             return null;
    //         }

    //         // FIXME: verify version more completely

    //         //
    //         return new InstallerGUI.MatlabInstallation(version, arch);
    //     }
    //     catch (IOException ex)
    //     {
    //         this.logger.warning("Unable to verify MATLAB installation in " + matlabPath + ": "
    //                             + ex.getMessage() + "\n");
    //         return null;
    //         // this.logger.severe("Unable to run MATLAB: " + ex.getMessage() + "\n");
    //         // return null;
    //     }
    //     catch (ProcessExecutionException ex)
    //     {
    //         this.logger.warning("Unable to verify MATLAB installation in " + matlabPath + ": "
    //                             + ex.getMessage() + "\n");
    //         return null;
    //         // this.logger.severe("Unable to run MATLAB: " + ex.getMessage() + "\n");
    //         // return null;
    //     }
    // }


    /*
     * return true if the superuser password is correct; false if not
     * FIXMEFIXME: unimplemented
     */
    protected boolean verifySuperuserPassword(Password superuserPassword)
    {
        // FIXME FILLME
        return false;
    }
}

