/*
 *
 */


/*
 *
 */
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.util.Vector;
import java.util.logging.Logger;

import installer.common.Controller;
import installer.common.MatlabRunner;
import installer.common.ProcessExecutionException;
import installer.common.ProcessOutputAnalyzer;
import installer.common.ProcessRunner;


/*
 * this is a singleton
 */
public class WindowsTempFileMatlabRunner extends MatlabRunner
{
    /*
     *
     */
    private static final WindowsTempFileMatlabRunner runner =
        new WindowsTempFileMatlabRunner();


    /*
     *
     */
    private WindowsTempFileMatlabRunner()
    {
    }


    /*
     * FIXME: this is an inherently unsafe method, but there is
     *        no suitable primitive to replace it, and we do
     *        need something like it
     */
    private static File createTempDirectory(String prefix,
                                            String suffix,
                                            File parentDirectory)
        throws IOException
    {
        File tempDir;


        //
        tempDir = File.createTempFile(prefix, suffix, parentDirectory);
        if (!tempDir.delete())
        {
            throw new IOException("Unable to delete temporary file: "
                                  + tempDir.getAbsolutePath());
        }
        if (!tempDir.mkdir())
        {
            throw new IOException("Unable to create temporary directory: "
                                  + tempDir.getAbsolutePath());
        }

        //
        return tempDir;
    }


    /*
     *
     */
    public static WindowsTempFileMatlabRunner get()
    {
        return runner;
    }


    /*
     * matlabCommand must not contain any semi-colons or
     * whitespace (most likely it will be just a single
     * function call without arguments)
     *
     * this doesn't actually use a temp. log file yet; it just
     * uses matlab.bat
     *
     * FIXMEFIXME: still need to make a temp. dir and so on
     *
     * this doesn't need all the machinery of ProcessRunner;
     * it just launches matlab with its gui
     *
     * FIXME: but, should figure out a way to tell when this has failed
     */
    public void launch(File matlabPath,
                       String matlabCommand,
                       File workingDirectory)
        throws IOException
    {
        String[] commandLine;
        File matlabBinaryPath;

        
        //
        matlabBinaryPath = new File(new File(matlabPath, "bin"), "matlab.bat");
        if (!matlabBinaryPath.exists())
        {
            throw new IOException("Expected MATLAB executable "
                                  + matlabBinaryPath.getAbsolutePath() + " "
                                  + "does not exist");
        }

        //
        commandLine =
            new String[] { matlabBinaryPath.getAbsolutePath(),
                           "-desktop",
                           "-r", matlabCommand };
        Runtime.getRuntime().exec(commandLine, null, workingDirectory);
    }


    /*
     *
     */
    public <ResultType> ResultType
        run(File matlabPath,
            String matlabCommand,
            File logFile,
            ProcessOutputAnalyzer<ResultType> outputAnalyzer)
        throws IOException,
               ProcessExecutionException
    {
        Logger logger;
        File matlabBinaryPath;
        Vector<String> commandLine;
        String[] commandLineArray;
        boolean logFileIsTemp;
        ResultType result;
        BufferedReader logFileReader;
        final String tempCommandName = "wfdbtmp";
        File tempDir, tempCommandFile;


        //
        logger = Logger.getLogger("controller");

        //
        matlabBinaryPath = new File(new File(matlabPath, "bin"), "matlab.bat");
        if (!matlabBinaryPath.exists())
        {
            throw new IOException("Expected MATLAB executable "
                                  + matlabBinaryPath.getAbsolutePath() + " "
                                  + "does not exist");
        }

        //
        logFileIsTemp = false;
        if (logFile == null && outputAnalyzer != null)
        {
            logFileIsTemp = true;
            logFile = File.createTempFile("wsm", null);
        }

        //
        commandLine = new Vector<String>();
        commandLine.add(matlabBinaryPath.getAbsolutePath());
        commandLine.add("-nodesktop");
        commandLine.add("-nosplash");
        commandLine.add("-wait");

        // NOTE: using -automation appears to interfere with the startup directory,
        //       which is critical for this method to work
        //commandLine.add("-automation");

        if (logFile != null)
        {
            commandLine.add("-logfile");
            commandLine.add(logFile.getAbsolutePath());
        }
        commandLine.add("-r");
        commandLine.add(tempCommandName);
        commandLineArray = commandLine.toArray(new String[0]);

        //
        logger.info("Running MATLAB (" + matlabBinaryPath.getAbsolutePath() + ") with "
                    + ((logFile != null)
                       ? ("log file " + logFile.getAbsolutePath() + " and ")
                       : "")
                    + "command: " + matlabCommand + " ...\n");

        //
        tempDir = createTempDirectory("wsm", null, null);

        //
        tempCommandFile = new File(tempDir, tempCommandName + ".m");
        try
        {
            BufferedWriter commandFileWriter;

            //
            if (!tempCommandFile.createNewFile())
            {
                throw new IOException("Unable to create file: "
                                      + tempCommandFile.getAbsolutePath() + "\n");
            }

            //
            commandFileWriter =
                new BufferedWriter(new OutputStreamWriter(new FileOutputStream(tempCommandFile)));
            commandFileWriter.write(matlabCommand + "\n" + "exit;\n");
            commandFileWriter.close();

            //
            ProcessRunner.runWithWorkingDirectory(null, tempDir, commandLineArray);

            //
            try
            {
                if (outputAnalyzer != null)
                {
                    logFileReader =
                        new BufferedReader(new InputStreamReader(new FileInputStream(logFile)));
                    try
                    {
                        result = outputAnalyzer.analyzeOutput(logFileReader);
                    }
                    catch (ProcessOutputAnalyzer.UnexpectedOutputException ex)
                    {
                        // this essentially mimics what ProcessRunner does
                        throw new ProcessExecutionException(commandLineArray, ex.getMessage());
                    }
                }
                else
                {
                    result = null;
                }
            }
            finally
            {
                if (logFile != null && logFileIsTemp && logFile.exists())
                {
                    if (!logFile.delete())
                    {
                        logger.warning("Unable to delete temporary file: "
                                       + logFile.getAbsolutePath() + " "
                                       + "(not a fatal error.)\n");
                    }
                }
            }
        }
        finally
        {
            try
            {
                Controller.removeDirectory(tempDir);
            }
            catch (IOException ex)
            {
                logger.warning("Unable to remove directory: "
                               + tempDir.getAbsolutePath() + " "
                               + "(not a fatal error.)\n");
            }
        }

        //
        return result;
    }
}