/*
 *
 */
//
//
// File: rdsampMatlab.java   Michael Craig  November 4 2008
// File: rdsamp.java       I. Henry    February 16 2005
//
// Minimal WFDB sample reader written Java, based on rdsamp.c
//
// Copyright (C) 2005 Isaac C. Henry (ihenry@physionet.org)
//
//

/*
 *
 */
import java.lang.Class;
import java.lang.ClassLoader;
import java.lang.reflect.InvocationTargetException;
import java.util.*;
import wfdb.*;


/*
 *
 */
public class rdsampMatlab
{
    /*
     *
     */
    static
    {
        try
        {
            Class nativeLibLoader
                = ClassLoader.getSystemClassLoader().loadClass("NativeLibLoader");
            nativeLibLoader.getMethod("loadLibraries", null).invoke(null, null);
        }
        catch (ClassNotFoundException ex)
        {
        }
        catch (NoSuchMethodException ex)
        {
        }
        catch (IllegalAccessException ex)
        {
        }
        catch (InvocationTargetException ex)
        {
        }
    }


    // /*
    //  * FIXME: this is debugging only!
    //  */
    // public static void main(String[] args)
    // {
    //     double[][] a = rdsamp("mitdb/100");
    //     int i;

    //     for (i = 0; i < a.length; ++i)
    //     {
    //         System.out.println(a[i][0] + " " + a[i][1] + " " + a[i][2]);
    //     }
    // }


    /*
     *
     */
    public static double[][] rdsamp(String recordName)
    {
        return rdsamp(recordName, null, null, null, false, false, null);
    }


    /*
     *
     */
    public static double[][] rdsamp(String recordName,
                                    String beginTime,
                                    String maxTimeInterval)
    {
        return rdsamp(recordName, beginTime, maxTimeInterval, null, false, false, null);
    }


    /*
     *
     * - the following command-line flags from rdsamp.c are ignored:
     *   -h, -p -p, -v
     *
     * - relevant command-line flags are shown next to their
     *   corresponding method arguments, below
     *
     * - if signalsToRead is null, then it is as if "-s" were
     *   not used
     *
     * - there are no column headers, as rdsamp.c would print out
     *
     * - if a sample is invalid, it becomes NaN
     *   (rdsamp.c would print "-" for it)
     *
     * - FIXME: test NaN/WFDB_INVALID_SAMPLE
     *
     */
    public static double[][] rdsamp(String recordName,        // -r
                                    String beginTime,         // -f
                                    String maxTimeInterval,   // -l
                                    String stopTime,          // -t
                                    boolean highRes,          // -H
                                    boolean usePhysicalUnits, // -p
                                    int[] signalsToRead)      // -s
    {
        int i;
        int nsig;
        int from, to, maxl;
        WFDB_SiginfoArray siarray;
        WFDB_SampleArray v;
        Vector outputRows;


        //
        nsig = wfdb.isigopen(recordName, null, 0);
        if (nsig <= 0)
        {
            System.out.println("error: unable to open record " + recordName);
            wfdb.wfdbquit();
            return null;
        }

        //
        v = new WFDB_SampleArray(nsig);
        siarray = new WFDB_SiginfoArray(nsig);
        nsig = wfdb.isigopen(recordName, siarray.cast(), nsig);
        if (nsig <= 0)
        {
            System.out.println("error: unable to open record " + recordName);
            wfdb.wfdbquit();
            return null;
        }

        // if a signal has zero gain specified, give it the default gain
        for (i = 0; i < nsig; ++i)
        {
            WFDB_Siginfo si = siarray.getitem(i);

            if (si.getGain() == 0.0)
            {
                si.setGain(wfdb.WFDB_DEFGAIN);
                siarray.setitem(i, si);
            }
        }

        //
        if (highRes)
        {
            wfdb.setgvmode(wfdb.WFDB_HIGHRES);
        }
        else
        {
            wfdb.setgvmode(wfdb.WFDB_LOWRES);
        }

        // establish the time interval for which to read the signal(s)
        from = 0;
        to = 0;
        maxl = 0;
        if (beginTime != null && (from = wfdb.strtim(beginTime)) < 0L)
        {
            from = -from;
        }
        if (wfdb.isigsettime(from) < 0)
        {
            System.out.println("error: invalid begin-time");
            wfdb.wfdbquit();
            return null;
        }
        if (stopTime != null && (to = wfdb.strtim(stopTime)) < 0L)
        {
            to = -to;
        }
        if (maxTimeInterval != null
            && (maxl = wfdb.strtim(maxTimeInterval)) < 0L)
        {
            maxl = -maxl;
        }
        if ((maxl != 0L) && (to == 0L || to > from + maxl))
        {
            to = from + maxl;
        }
        
        //
        if (signalsToRead != null) // print samples only from specified signals
        {
            if (signalsToRead.length == 0)
            {
                System.out.println("error: the signals-to-read array is non-null, but empty");
                wfdb.wfdbquit();
                return null;
            }

            for (i = 0; i < signalsToRead.length; ++i)
            {
                if (signalsToRead[i] < 0 || signalsToRead[i] >= nsig)
                {
                    System.out.println("error: invalid entry in signals-to-read array");
                    wfdb.wfdbquit();
                    return null;
                }
            }
        }
        else // print samples from all signals
        {
            signalsToRead = new int[nsig];
            for (i = 0; i < nsig; ++i)
            {
                signalsToRead[i] = i;
            }
        }

        //
        outputRows = new Vector();

        //
        if (usePhysicalUnits)
        {
            double freq;
            double[] baselines, gains;

            
            // "cache" the baselines and gains -- do not force a call to
            // WFDB_SiginfoArray.getitem() every time we need them.
            baselines = new double[signalsToRead.length];
            gains = new double[signalsToRead.length];
            for (i = 0; i < signalsToRead.length; ++i)
            {
                WFDB_Siginfo si = siarray.getitem(signalsToRead[i]);
                baselines[i] = si.getBaseline();
                gains[i] = si.getGain();
            }

            freq = wfdb.sampfreq(null);

            while ((to == 0L || from < to) && wfdb.getvec(v.cast()) >= 0)
            {
                double[] row = new double[signalsToRead.length + 1];
                int d;

                for (i = 0, row[0] = ((double)from)/freq;
                     i < signalsToRead.length;
                     ++i)
                {
                    d = v.getitem(signalsToRead[i]);
                    if (d != wfdb.WFDB_INVALID_SAMPLE)
                    {
                        // row[i + 1] =
                        //     ((double)(v.getitem(signalsToRead[i]) -
                        //               siarray.getitem(signalsToRead[i]).getBaseline())) /
                        //     siarray.getitem(signalsToRead[i]).getGain();
                        row[i + 1] = ((double)(d - baselines[i])) / gains[i];
                    }
                    else
                    {
                        row[i + 1] = Double.NaN;
                    }
                }

                outputRows.add(row);
                ++from;
            }
        }
        else
        {
            while ((to == 0L || from < to) && wfdb.getvec(v.cast()) >= 0)
            {
                double[] row = new double[signalsToRead.length + 1];

                for (i = 0, row[0] = from; i < signalsToRead.length; ++i)
                {
                    row[i + 1] = v.getitem(signalsToRead[i]);
                }
                
                outputRows.add(row);
                ++from;
            }
        }

        //
        wfdb.wfdbquit();

        //
        return (double[][]) outputRows.toArray(new double[outputRows.size()][]);
    }
}
